/*	                     __      __
	                     \ \    / /
	                      \ \  / /
	         _      _______\ \/ /_______      _
	        / \    / __________________ \    / \
	       (   \  | /        __        \ |  /   )
	     ___\  |  | |       / _\       | |  |  /___
	    |____  |  | |      ( <_        | |  | _____|
	      / \| |  | |       \__/       | |  | |/ \
	     / \ | |  | \__________________/ |  | | / \
	    /_/ \__/   \____________________/   \__/ \_\
	 ___  _   _ ___ __  __   _   _  _ ___   _     ____ __
	|   \| | | / __|  \/  | /_\ | \| |_ _| /_\   |__  /  \
	| |) | |_| \__ \ |\/| |/ _ \| .` || | / _ \    / / () |
	|___/ \___/|___/_|  |_/_/ \_\_|\_|___/_/ \_\  /_(_)__/

	      Framework fr den Spieleprogrammierwettbewerb
	             by David Scherfgen (TomasRiker)
*/

#define INITGUID
#include <cassert>
#include "safe.h"
#include "input.h"
#include "graphics.h"
using namespace std;

namespace FrameWork
{
	Input::Input()
	{
		assert(Graphics::isInitialized());

		// DirectInput initialisieren
		HRESULT hres = DirectInput8Create(GetModuleHandle(0), DIRECTINPUT_VERSION, IID_IDirectInput8,
										  reinterpret_cast<void**>(&p_dInput), 0);
		assert(SUCCEEDED(hres));
		assert(p_dInput);

		// Tastatur initialisieren
		hres = p_dInput->CreateDevice(GUID_SysKeyboard, &p_keyboard, 0);
		assert(SUCCEEDED(hres));
		assert(p_keyboard);

		HWND window = Graphics::inst().getWindow();
		hres = p_keyboard->SetCooperativeLevel(window, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE);
		assert(SUCCEEDED(hres));
		hres = p_keyboard->SetDataFormat(&c_dfDIKeyboard);
		assert(SUCCEEDED(hres));

		DIPROPDWORD dipdw;
		dipdw.diph.dwSize = sizeof(DIPROPDWORD);
		dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER);
		dipdw.diph.dwObj = 0;
		dipdw.diph.dwHow = DIPH_DEVICE;
		dipdw.dwData = 256;
		hres = p_keyboard->SetProperty(DIPROP_BUFFERSIZE, &dipdw.diph);
		assert(SUCCEEDED(hres));

		// Maus initialisieren
		hres = p_dInput->CreateDevice(GUID_SysMouse, &p_mouse, 0);
		assert(SUCCEEDED(hres));
		assert(p_mouse);
		hres = p_mouse->SetCooperativeLevel(window, DISCL_FOREGROUND | DISCL_EXCLUSIVE);
		assert(SUCCEEDED(hres));
		hres = p_mouse->SetDataFormat(&c_dfDIMouse2);
		assert(SUCCEEDED(hres));

		dipdw.diph.dwSize = sizeof(DIPROPDWORD);
		dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER);
		dipdw.diph.dwObj = 0;
		dipdw.diph.dwHow = DIPH_DEVICE;
		dipdw.dwData = 256;
		hres = p_mouse->SetProperty(DIPROP_BUFFERSIZE, &dipdw.diph);
		assert(SUCCEEDED(hres));

		acquireDevices();

		ZeroMemory(keyboardKey, sizeof(keyboardKey));
		ZeroMemory(mouseButton, sizeof(mouseButton));
		ZeroMemory(mouseAxis, sizeof(mouseAxis));
	}

	Input::~Input()
	{
		safeRelease(p_dInput);
	}

	void Input::acquireDevices()
	{
		p_keyboard->Acquire();
		p_mouse->Acquire();
	}

	void Input::unacquireDevices()
	{
		p_keyboard->Unacquire();
		p_mouse->Unacquire();
	}

	vector<InputEvent> Input::processInput()
	{
		vector<InputEvent> events;

		for(int i = 0; i < 256; i++) keyboardKey[i] &= 1;
		for(int j = 0; j < 8; j++) mouseButton[j] &= 1;
		ZeroMemory(mouseAxis, sizeof(mouseAxis));

		// Tastaturereignisse verarbeiten
		DIDEVICEOBJECTDATA data;
		while(true)
		{
			DWORD n = 1;
			HRESULT hres = p_keyboard->GetDeviceData(sizeof(DIDEVICEOBJECTDATA), &data, &n, 0);
			if(FAILED(hres) || n == 0) break;

			assert(data.dwOfs > 0 && data.dwOfs < 256);

			if(data.dwData & 0x80)
			{
				// Taste gedrckt
				InputEvent evt;
				evt.type = IET_KEYBOARD_KEY_DOWN;
				evt.key = data.dwOfs;
				events.push_back(evt);
				keyboardKey[evt.key] |= 1;
				keyboardKey[evt.key] |= 2;
			}
			else
			{
				// Taste losgelassen
				InputEvent evt;
				evt.type = IET_KEYBOARD_KEY_UP;
				evt.key = data.dwOfs;
				events.push_back(evt);
				keyboardKey[evt.key] &= ~1;
				keyboardKey[evt.key] |= 4;
			}
		}

		// Mausereignisse verarbeiten
		while(true)
		{
			DWORD n = 1;
			HRESULT hres = p_mouse->GetDeviceData(sizeof(DIDEVICEOBJECTDATA), &data, &n, 0);
			if(FAILED(hres) || n == 0) break;

			if(data.dwOfs >= DIMOFS_BUTTON0 && data.dwOfs <= DIMOFS_BUTTON7)
			{
				// Mausknopf gedrckt oder losgelassen
				int button = data.dwOfs - DIMOFS_BUTTON0;
				assert(button >= 0 && button < 8);

				if(data.dwData & 0x80)
				{
					// Mausknopf gedrckt
					InputEvent evt;
					evt.type = IET_MOUSE_BUTTON_DOWN;
					evt.button = button;
					events.push_back(evt);
					mouseButton[evt.button] |= 1;
					mouseButton[evt.button] |= 2;
				}
				else
				{
					// Mausknopf losgelassen
					InputEvent evt;
					evt.type = IET_MOUSE_BUTTON_UP;
					evt.button = button;
					events.push_back(evt);
					mouseButton[evt.button] &= ~1;
					mouseButton[evt.button] |= 4;
				}
			}
			else if(data.dwOfs >= DIMOFS_X && data.dwOfs <= DIMOFS_Z)
			{
				// Es befand eine Bewegung statt.
				int axis = (data.dwOfs - DIMOFS_X) / sizeof(long);
				assert(axis >= 0 && axis < 3);

				InputEvent evt;
				evt.type = IET_MOUSE_MOVEMENT;
				ZeroMemory(&evt.movement, sizeof(evt.movement));
				evt.movement[axis] = static_cast<int>(data.dwData);
				events.push_back(evt);
				mouseAxis[axis] += evt.movement[axis];
			}
		}

		return events;
	}
}
