/*	                     __      __
	                     \ \    / /
	                      \ \  / /
	         _      _______\ \/ /_______      _
	        / \    / __________________ \    / \
	       (   \  | /        __        \ |  /   )
	     ___\  |  | |       / _\       | |  |  /___
	    |____  |  | |      ( <_        | |  | _____|
	      / \| |  | |       \__/       | |  | |/ \
	     / \ | |  | \__________________/ |  | | / \
	    /_/ \__/   \____________________/   \__/ \_\
	 ___  _   _ ___ __  __   _   _  _ ___   _     ____ __
	|   \| | | / __|  \/  | /_\ | \| |_ _| /_\   |__  /  \
	| |) | |_| \__ \ |\/| |/ _ \| .` || | / _ \    / / () |
	|___/ \___/|___/_|  |_/_/ \_\_|\_|___/_/ \_\  /_(_)__/

	      Framework fr den Spieleprogrammierwettbewerb
	             by David Scherfgen (TomasRiker)
*/

#include <cassert>
#include "safe.h"
#include "particlesystem.h"
#include "particleemitter.h"
#include "surface.h"
using namespace std;

namespace FrameWork
{
	ParticleSystem::ParticleSystem(Surface* p_sprites)
	{
		this->p_sprites = p_sprites;
		gravity = D3DXVECTOR2(0.0f, 98.1f);
	}

	ParticleSystem::~ParticleSystem()
	{
		deleteAllEmitters();
	}

	ParticleEmitter* ParticleSystem::createEmitter(const ivec2& spriteStartOnTex,
												   const ivec2& spriteSizeOnTex)
	{
		ParticleEmitter* p_emitter = new ParticleEmitter(spriteStartOnTex, spriteSizeOnTex);
		assert(p_emitter);
		emitters.push_back(p_emitter);
		return p_emitter;
	}

	void ParticleSystem::deleteEmitter(ParticleEmitter* p_emitter)
	{
		assert(p_emitter);
		delete p_emitter;
		emitters.remove(p_emitter);
	}

	void ParticleSystem::deleteAllEmitters()
	{
		while(!emitters.empty()) deleteEmitter(*emitters.begin());
	}

	bool ParticleSystem::isValidEmitter(const ParticleEmitter* p_emitter) const
	{
		for(list<ParticleEmitter*>::const_iterator i = emitters.begin(); i != emitters.end(); i++) if(*i == p_emitter) return true;
		return false;
	}

	void ParticleSystem::render(Surface* p_dest)
	{
		assert(p_dest);

		// die Partikel rendern
		for(list<Particle>::iterator i = particles.begin(); i != particles.end(); i++)
		{
			Particle& p = *i;
			p_dest->drawSubSprite(p_sprites, p.spriteStartOnTex, p.spriteSizeOnTex,
								  p.position, D3DXVECTOR2(p.size, p.size), p.orientation, p.color);
		}
	}

	void ParticleSystem::update(float dt)
	{
		// die Partikel aktualisieren
		for(list<Particle>::iterator i = particles.begin(); i != particles.end();)
		{
			Particle& p = *i;
			p.lifetimeLeft -= dt;
			p.position += dt * p.velocity;
			p.color += dt * p.deltaColor;
			p.orientation += dt * p.rotation;
			p.size += dt * p.deltaSize;

			if(p.gravityInfluence != 0.0f) p.velocity += dt * gravity * p.gravityInfluence;
			if(p.damping != 1.0f) p.velocity *= powf(p.damping, dt);

			if(p.lifetimeLeft < 0.0f) i = particles.erase(i);
			else i++;
		}

		// die Emitter neue Partikel erzeugen lassen
		for(list<ParticleEmitter*>::iterator j = emitters.begin(); j != emitters.end();)
		{
			ParticleEmitter* p_emitter = *j;
			if(p_emitter->toBeDestroyed())
			{
				// Der automatische Zerstrungs-Count-Down ist abgelaufen!
				delete p_emitter;
				j = emitters.erase(j);
			}
			else
			{
				p_emitter->emit(particles, dt);
				j++;
			}
		}
	}
}
