/*	                     __      __
	                     \ \    / /
	                      \ \  / /
	         _      _______\ \/ /_______      _
	        / \    / __________________ \    / \
	       (   \  | /        __        \ |  /   )
	     ___\  |  | |       / _\       | |  |  /___
	    |____  |  | |      ( <_        | |  | _____|
	      / \| |  | |       \__/       | |  | |/ \
	     / \ | |  | \__________________/ |  | | / \
	    /_/ \__/   \____________________/   \__/ \_\
	 ___  _   _ ___ __  __   _   _  _ ___   _     ____ __
	|   \| | | / __|  \/  | /_\ | \| |_ _| /_\   |__  /  \
	| |) | |_| \__ \ |\/| |/ _ \| .` || | / _ \    / / () |
	|___/ \___/|___/_|  |_/_/ \_\_|\_|___/_/ \_\  /_(_)__/

	      Framework fr den Spieleprogrammierwettbewerb
	             by David Scherfgen (TomasRiker)
*/

#include <cassert>
#include <cstdio>
#include "safe.h"
#include "sound.h"
#include "soundinstance.h"
#include "audio.h"
#include "utils.h"
using namespace std;

namespace FrameWork
{
	list<Sound*> Sound::sounds;

	// RIFF-Header
	struct RIFFHeader
	{
		char riff[4];
		DWORD size;
		char format[4];
	};

	// Header eines WAV-Datei-Chunks
	struct WAVChunkHeader
	{
		char type[4];
		DWORD dataSize;
	};

	IDirectSoundBuffer* loadSound(const string& filename)
	{
		assert(Audio::isInitialized());

		// Datei ffnen
		FILE* p_file = fopen(filename.c_str(), "rb");
		if(!p_file) return 0;

		// Header lesen und sicherstellen, dass es wirklich eine WAV-Datei ist
		RIFFHeader riffHeader;
		fread(&riffHeader, sizeof(riffHeader), 1, p_file);
		if(strnicmp(riffHeader.riff, "RIFF", 4) || strnicmp(riffHeader.format, "WAVE", 4))
		{
			fclose(p_file);
			return 0;
		}

		WAVEFORMATEX waveFormat;
		char* p_soundData = 0;
		int soundDataSize = 0;
		bool formatChunkRead = false;
		bool dataChunkRead = false;

		while(true)
		{
			// Chunk-Header lesen
			WAVChunkHeader chunkHeader;
			if(fread(&chunkHeader, sizeof(chunkHeader), 1, p_file) != 1) break;

			// Format-Chunk?
			if(!strnicmp(chunkHeader.type, "fmt ", 4))
			{
				// Format lesen
				DWORD numBytesToRead = dus_min<unsigned int>(chunkHeader.dataSize, sizeof(WAVEFORMATEX));
				fread(&waveFormat, numBytesToRead, 1, p_file);
				formatChunkRead = true;
			}
			// Daten-Chunk?
			else if(!strnicmp(chunkHeader.type, "data", 4))
			{
				// Daten lesen
				soundDataSize = chunkHeader.dataSize;
				p_soundData = new char[soundDataSize];
				assert(p_soundData);
				fread(p_soundData, soundDataSize, 1, p_file);
				dataChunkRead = true;
			}
			else
			{
				// Chunk berspringen
				fseek(p_file, chunkHeader.dataSize, SEEK_CUR);
			}
		}

		// Wir brauchen einen Format- und einen Daten-Chunk!
		if(!(formatChunkRead && dataChunkRead))
		{
			safeDelete(p_soundData);
			return 0;
		}

		// Pufferbeschreibung ausfllen
		DSBUFFERDESC bufferDesc;
		ZeroMemory(&bufferDesc, sizeof(bufferDesc));
		bufferDesc.dwSize = sizeof(bufferDesc);
		bufferDesc.dwFlags = DSBCAPS_STATIC | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLPAN | DSBCAPS_CTRLFREQUENCY;
		bufferDesc.dwBufferBytes = soundDataSize;
		bufferDesc.dwReserved = 0;
		bufferDesc.lpwfxFormat = &waveFormat;

		// Soundpuffer erstellen
		IDirectSoundBuffer* p_buffer;
		HRESULT hres = Audio::inst().getDSound()->CreateSoundBuffer(&bufferDesc, &p_buffer, 0);
		assert(SUCCEEDED(hres));
		assert(p_buffer);

		// Puffer sperren und Daten hineinkopieren
		void* p_soundBufferData;
		DWORD soundBufferDataSize;
		p_buffer->Lock(0, 0, &p_soundBufferData, &soundBufferDataSize, 0, 0, DSBLOCK_ENTIREBUFFER);
		memcpy(p_soundBufferData, p_soundData, dus_min<unsigned int>(soundDataSize, static_cast<int>(soundBufferDataSize)));
		p_buffer->Unlock(p_soundBufferData, soundBufferDataSize, 0, 0);
		safeDelete(p_soundData);

		return p_buffer;
	}

	Sound::Sound(const string& filename)
	{
		assert(Audio::isInitialized());

		p_buffer = loadSound(filename);
		assert(p_buffer);

		p_dSound = Audio::inst().getDSound();
		p_dSound->AddRef();

		sounds.push_back(this);
	}

	Sound::Sound(HMODULE module,
				 LPCTSTR p_resource)
	{
		assert(Audio::isInitialized());

		HRSRC resInfo = FindResource(module, p_resource, "RT_RCDATA");
		assert(resInfo);
		int resSize = SizeofResource(module, resInfo);
		HGLOBAL res = LoadResource(module, resInfo);
		assert(res);
		void* p_resData = LockResource(res);
		assert(p_resData);

		FILE* p_file = fopen("___temp.dat", "wb");
		assert(p_file);
		fwrite(p_resData, resSize, 1, p_file);
		fclose(p_file);

		UnlockResource(res);
		FreeResource(res);

		p_buffer = loadSound("___temp.dat");
		assert(p_buffer);

		p_dSound = Audio::inst().getDSound();
		p_dSound->AddRef();

		DeleteFile("___temp.dat");

		sounds.push_back(this);
	}

	Sound::~Sound()
	{
		while(!instances.empty()) deleteInstance(*instances.begin());
		safeRelease(p_buffer);
		safeRelease(p_dSound);
		sounds.remove(this);
	}

	SoundInstance* Sound::createInstance(bool persist)
	{
		for(list<Sound*>::iterator i = sounds.begin(); i != sounds.end(); i++) (*i)->gc();
		SoundInstance* p_inst = new SoundInstance(p_buffer, persist);
		instances.push_back(p_inst);
		return p_inst;
	}

	void Sound::deleteInstance(SoundInstance* p_inst)
	{
		assert(p_inst);
		delete p_inst;
		instances.remove(p_inst);
	}

	bool Sound::isValidInstance(const SoundInstance* p_inst)
	{
		for(list<SoundInstance*>::const_iterator i = instances.begin(); i != instances.end(); i++) if(*i == p_inst) return true;
		return false;
	}

	void Sound::gc()
	{
		for(list<SoundInstance*>::iterator i = instances.begin(); i != instances.end();)
		{
			// Sound evtl. freigeben
			if((*i)->toBeDeleted())
			{
				delete *i;
				i = instances.erase(i);
			}
			else i++;
		}
	}
}
